/* *******************************************************************
   NTP (SNTP)

   to sync the internal time (from timelib.h) with NTP we use
   - timerNTP()           a timer to check if it's time to call NTP
   - sendNTPpacket()      the UDP request to the NTP Server
   - checkIncommingUdp()  a collector of incomming UDP Packets and update of internal time
   ***************************************************************** */

void timerNTP(bool forced)             // if forced=true the internal timer will be ignored
{
  // for the NTP we use a small FSM
  enum State {IDLE,                    // only check if it is time to initiate a new NTP call
              WAIT_FOR_RESPONSE
             };      // check if we got a new NTP response
  static State state = IDLE;

  uint32_t interval = 1 * 60 * 1000UL;                                         // till first successfull set: call each minute
  static uint32_t previousMillis = 0 - interval + random(4000, 15000);         // first NTP check is 4..15 seconds after boot
  if (timeStatus() != timeNotSet) interval = 60 * 60 * 1000UL;                 // if time is set once, renew NTP every hour

  // check if it is time for NTP
  if ((state == IDLE && millis() - previousMillis > interval) || forced)
  {
    previousMillis = millis();
    state = WAIT_FOR_RESPONSE;
    while (Udp.parsePacket() > 0) ;    // discard any previously received packets - similar to time.h example
    sendNTPpacket(timeServer);         // send an NTP packet to a time server
  }
  if (state == WAIT_FOR_RESPONSE)
  {
    if (checkIncommingUdp() == 0)      // check if we received some UDP packet
      state = IDLE;
    if (millis() - previousMillis > 1500)
    {
      Serial.println(F("NTP Timeout - no answer"));
      state = IDLE;
    }
  }
}


// send a NTP request to the time server at the given address
void sendNTPpacket(const char * address) {
  Serial.println(F("send NTP packet"));
  memset(packetBuffer, 0, NTP_PACKET_SIZE);      // set all bytes in the buffer to 0
  // Initialize values needed to form NTP request
  packetBuffer[0] = 0b11100011;        // LI, Version, Mode
  packetBuffer[1] = 0;                 // Stratum, or type of clock
  packetBuffer[2] = 6;                 // Polling Interval
  packetBuffer[3] = 0xEC;              // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12]  = 49;
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123);       // NTP requests are to port 123
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}


int checkIncommingUdp()
{
  int result = -1;       // return errorcode
  int size = Udp.parsePacket();
  if (size)
    //if (Udp.parsePacket())
  {
    Serial.print(F("size=")); Serial.println(size);
    // We've received a packet, read the data from it
    Udp.read(packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer
    // the timestamp starts at byte 40 of the received packet and is four bytes,
    // or two words, long. First, extract the two words:
    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
    // combine the four bytes (two words) into a long integer
    // this is NTP time (seconds since Jan 1 1900):
    unsigned long secsSince1900 = highWord << 16 | lowWord;
    Serial.print(F("Seconds since Jan 1 1900 = "));
    Serial.println(secsSince1900);
    // now convert NTP time into everyday time:
    Serial.print(F("Unix time = "));
    // Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
    const unsigned long seventyYears = 2208988800UL;
    // subtract seventy years:
    unsigned long epoch = secsSince1900 - seventyYears;
    // print Unix time:
    Serial.println(epoch);
    // print the hour, minute and second:
    Serial.print(F("The UTC time is "));         // UTC is the time at Greenwich Meridian (GMT)
    Serial.print((epoch  % 86400L) / 3600);      // print the hour (86400 equals secs per day)
    Serial.print(':');
    if (((epoch % 3600) / 60) < 10) {
      // In the first 10 minutes of each hour, we'll want a leading '0'
      Serial.print('0');
    }
    Serial.print((epoch  % 3600) / 60);          // print the minute (3600 equals secs per minute)
    Serial.print(':');
    if ((epoch % 60) < 10) {
      // In the first 10 seconds of each minute, we'll want a leading '0'
      Serial.print('0');
    }
    Serial.println(epoch % 60);        // print the second
    // handover time to timelib
    setTime(epoch);          // use the epoch, not the secSince1900
    result = 0;
  }
  return result;
}


/* *****************************
   unused helper functions
 * **************************** */

void timerClock()
{
  static uint32_t previousMillis = -1000;
  if (millis() - previousMillis > 1000)
  {
    previousMillis = millis();
    Serial.print(hour());
    Serial.print(":");
    print00 (minute());
    Serial.print(":");
    print00 (second());
    Serial.print(" ");
    Serial.print(year());
    Serial.print("-");
    print00(month());
    Serial.print("-");
    print00(day());
    if (timeStatus() == timeNotSet) Serial.print(F(" not set")); else Serial.print(F(" set"));
    Serial.println();
  }
}

// helper function to print digits with pre 0
void print00(int actual)
{
  if (actual < 10) Serial.print("0");
  Serial.print(actual);
}
